struct PCA
    Scores::Array{Float64,2}
    Loadings::Array{Float64,2}
    Values::Array
    algorithm::String
end

"""
    PCA_NIPALS(X; Factors = minimum(size(X)) - 1, tolerance = 1e-7, maxiters = 200)

Compute's a PCA from `x` using the NIPALS algorithm with a user specified number of latent variables(`Factors`).
The tolerance is the minimum change in the F norm before ceasing execution. Returns a PCA object.

"""
function PCA_NIPALS(X; Factors = minimum(size(X)) - 1, tolerance = 1e-7, maxiters = 200)
    tolsq = tolerance * tolerance
    #Instantiate some variables up front for performance...
    Xsize = size(X)
    Tm = zeros( ( Xsize[1], Factors ) )
    Pm = zeros( ( Factors, Xsize[2] ) )
    t = zeros( ( 1, Xsize[1] ) )
    p = zeros( ( 1, Xsize[2] ) )
    #Set tolerance to floating point precision
    Residuals = copy(X)
    for factor in 1:Factors
        lastErr = sum(abs.(Residuals)); curErr = tolerance + 1;
        t = Residuals[:, 1]
        iterations = 0
        while (abs(curErr - lastErr) > tolsq) && (iterations < maxiters)
            p = Residuals' * t
            p = p ./ sqrt.( p' * p )
            t = Residuals * p
            #Track change in Frobenius norm
            lastErr = curErr
            curErr = sqrt(sum( ( Residuals - ( t * p' ) ) .^ 2))
            iterations += 1
        end
        Residuals -= t * p'
        Tm[:,factor] = t
        Pm[factor,:] = p
    end
    #Find singular values/eigenvalues
    EigVal = sqrt.( LinearAlgebra.diag( Tm' * Tm ) )
    #Scale loadings by singular values
    Tm = Tm * LinearAlgebra.Diagonal( 1.0 / EigVal )
    return PCA(Tm, Pm, EigVal, "NIPALS")
end

"""
    PCA(X; Factors = minimum(size(X)) - 1)

Compute's a PCA from `x` using LinearAlgebra's SVD algorithm with a user specified number of latent variables(`Factors`).
Returns a PCA object.

"""
function PCA(Z; Factors = minimum(size(Z)) - 1)
    svdres = LinearAlgebra.svd(Z)
    return PCA(svdres.U[:, 1:Factors], svdres.Vt[1:Factors, :], svdres.S[1:Factors], "SVD")
end

"""
    (T::PCA)(Z::Array; Factors = length(T.Values), inverse = false)

Calling a PCA object on new data brings the new data `Z` into or out of (`inverse` = true) the PCA basis.

"""
function (T::PCA)(Z::Array; Factors = length(T.Values), inverse = false)
    if (inverse)
        return Z * (Diagonal(T.Values[1:Factors]) * T.Loadings[1:Factors,:])
    else
        return Z * (Diagonal( 1.0 ./ T.Values[1:Factors]) * T.Loadings[1:Factors,:])'
    end
end

struct LDA
    Scores::Array{Float64,2}
    Loadings::Array{Float64,2}
    Values::Array
end

"""
    LDA(X, Y; Factors = 1)

Compute's a LinearDiscriminantAnalysis transform from `x` with a user specified number of latent variables(`Factors`).
Returns an LDA object.
"""
function LDA(X, Y; Factors = 1)
    (Obs, ClassNumber) = size( Y )
    @assert Factors < ClassNumber
    Variables = size( X )[ 2 ]
    #Instantiate some variables...
    ClassMeans = zeros( ClassNumber, Variables )
    ClassSize = zeros( ClassNumber )
    WithinCovariance = zeros( Variables, Variables )
    BetweenCovariance = zeros( Variables, Variables  )
    ClassCovariance = zeros( Variables, Variables  )
    for class in 1 : ClassNumber
        Members = Y[ :, class ] .== 1
        ClassSize[class] = sum(Members)
        ClassMeans[class,:] = StatsBase.mean(X[Members,:], dims = 1)
    end
    GlobalMean = StatsBase.mean(X, dims = 1)

    for class in 1 : ClassNumber
        Members = Y[ :, class ] .== 1
        #calculate the between class covariance matrix
        Diff = ClassMeans[class,:] .- GlobalMean'
        BetweenCovariance .+= ClassSize[class] .* ( Diff * Diff' )
        #calculate the within class covariance matrix
        for member in findall(Members .== true)
            MeanCentered = X[member,:] .- ClassMeans[class, : ]
            WithinCovariance .+= ( MeanCentered * MeanCentered' )
        end
    end
    #Calculate the discriminant axis'
    #eig = LinearAlgebra.eigen(Base.inv(WithinCovariance) * BetweenCovariance)
    eig = LinearAlgebra.eigen(WithinCovariance \ BetweenCovariance)
    if any( imag.( eig.values ) .> 1e-1)
        println("Warning: Some eigenvalues found to have complex contributions > 0.1")
    end
    #Maybe reccomend to the user to do pca first or centerscale or both?
    ReVals = real.(eig.values)
    Sorted = sortperm( ReVals, rev = true)
    Contributions = ReVals[Sorted] .>= 1e-9
    Loadings = real.(eig.vectors[:,Sorted[ Contributions ][1:Factors] ] )
    #Project the X data into the LDA basis
    Scores = X * Loadings
    return LDA( Scores, Loadings, ReVals[ Sorted[ Contributions][1:Factors] ])
end

"""
    ( model::LDA )( Z; Factors = length(model.Values) )

Calling a LDA object on new data brings the new data `Z` into the LDA basis.

"""
function ( model::LDA )( Z; Factors = length(model.Values) )
     Projected = Z * model.Loadings[:,1:Factors]
end


"""
    HLDA(X, YHOT; K = 1, Factors = 1)

Compute's a Hierarchical LinearDiscriminantAnalysis transform from `x` with a user
specified number of latent variables(`Factors`). The adjacency matrices are created
from `K` nearest neighbors.

Returns an LDA object.
*Note: this can be used with any other LDA functions such as Gaussian discriminants
or explained variance*.

Lu D, Ding C, Xu J, Wang S. Hierarchical Discriminant Analysis. Sensors (Basel).
2018 Jan 18;18(1). pii: E279. doi: 10.3390/s18010279.
"""
function HLDA(X, YHOT; K = 1, Factors = 2)
    Dist = SquareEuclideanDistance( X )
    ICAM = OutOfClassAdjacencyMatrix(Dist, YHOT, K)
    ICD = LinearAlgebra.Diagonal(sum( ICAM, dims = 1) )
    ICS = ICD .- ICAM

    Decomp = eigen(X' * ICS * X)
    ReVals = real.(Decomp.values)
    Sorted = sortperm( ReVals, rev = false)
    Loadings = real.(Decomp.vectors[ : , Sorted[ 1:Factors ] ] )

    Dist = SquareEuclideanDistance( X * Loadings[:,1:Factors] )
    OCAM = InClassAdjacencyMatrix(Dist, YHOT, K)
    OCD = LinearAlgebra.Diagonal(sum( OCAM, dims = 1) )
    OCS = OCD .- OCAM

    Decomp = eigen(X' * OCS * X)
    ReVals = real.(Decomp.values)
    Sorted = sortperm( ReVals, rev = true)
    Contributions = ReVals[Sorted] .>= 1e-16
    if Factors > sum(Contributions)
        Factors = sum(Contributions)
    end
    Loadings = real.(Decomp.vectors[ : , Sorted[ Contributions ][ 1 : Factors ] ] )

    return LDA( X * Loadings,  Loadings, ReVals[ Sorted[ Contributions][1:Factors] ] )
end

#Untested...
struct CanonicalCorrelationAnalysis
    U
    V
    r
end

"""
    CanonicalCorrelationAnalysis(A, B)

Returns a CanonicalCorrelationAnalysis object which contains (U, V, r) from Arrays A and B.
Currently Untested for correctness but should compute....
"""
function CanonicalCorrelationAnalysis(A, B)
    (Obs,Vars) = size(A);;
    CAA = (1/Obs) .* A * A'
    CBB = (1/Obs) .* B * B'
    CAB = (1/Obs) .* A * B'
    maxrank = min( LinearAlgebra.rank( A ), LinearAlgebra.rank( B ) )
    CAAInvSqrt = CAA ^ -0.5
    CBBInvSqrt = CBB ^ -0.5
    singvaldecomp = LinearAlgebra.svd( CAAInvSqrt * CAB * CBBInvSqrt )
    Aprime = CAAInvSqrt * singvaldecomp.U[ :,1 : maxrank ]
    Bprime = CAAInvSqrt * singvaldecomp.Vt[ :,1 : maxrank ]
    return CanonicalCorrelationAnalysis(Aprime' * A, Bprime' * B, singvaldecomp.S[1 : maxrank] )
end

"""
    findpeaks( vY; m = 3)

Finds the indices of peaks in a vector vY with a window span of `2m`.
Original R function by Stas_G:(https://stats.stackexchange.com/questions/22974/how-to-find-local-peaks-valleys-in-a-series-of-data)
This version is based on a C++ variant by me.
"""
function findpeaks( vY; m = 3)
    @assert length(size(vY)) == 1
    sze = size(vY)[1];
    (i,q) = (0,0);#generic iterator, second generic iterator
    (lb,rb) = (0,0);#left bound, right bound
    ret = [];
    for i in 1 : ( sze - 2 )
        #Find all regions with negative laplacian between neighbors
        if (sign( vY[ i + 2 ]  - vY[ i + 1 ] ) - sign( vY[ i + 1 ]  - vY[ i ] ) ) < 0
            #Now assess all regions with negative laplacian between neighbors...
            lb = i - m - 1;# define left bound of vector
            lb = (lb < 1) ? 1 : lb
            rb = i + m + 1# define right bound of vector
            rb = (rb >= (sze-2)) ? rb = (sze-3) : rb
            #We have found a peak by our criterion
            if !any(vY[lb:rb] .> vY[i+1])
                push!( ret, i + 1 );
            end #End if found peak
        end #End if laplace condition
    end #End loop
    return ret
end

function RecursiveAlignment(aligned, reference; maxlags::Int = 800, lookahead::Int = 0, minlength::Int = 20, mincorr::Float64 = 0.05)
    if length(aligned) < minlength
        return aligned;
    end
    #Calculate cross correlation between the spectrum and the reference
    lagspan = min( length( aligned ) - 1, maxlags )
    r = StatsBase.crosscor( aligned, reference, -lagspan : lagspan  )
    maxi = argmax( r ) #Find largest value
    if r[ maxi ] <= mincorr # is the largest correlation junk?
        return aligned
    end
    lag = (maxi > lagspan/2) ? maxi - lagspan - 1 : maxi - 1;

    if lag == 0
        if lookahead <= 0
            return aligned
        end
        lookahead -= 1
    end

    #Pad spectra based on lags
    if !( ( lag == 0 ) || ( lag >= length( aligned ) ) )
        if lag > 0
    	   aligned = vcat( ones( lag ) * aligned[ 1 ], aligned[ 1 : (length(aligned) - lag) ] );
        elseif lag < 0
    	   lag = abs( lag );
    	   aligned = vcat( aligned[ ( lag + 1 ) : length( aligned ) ], ones( lag ) * aligned[ end ] );
        end
    end
    #Find new branch point
    middle = round( length( aligned ) / 2) |> Int
    quarter = floor( middle / 4 ) |> Int
    mini = argmin( aligned[ (middle - quarter ) : ( middle + quarter ) ] ) |> Int
    midpnt = mini + middle - quarter
    #Calculate branched components
    LHS = RecursiveAlignment(aligned[ 1 : midpnt ], reference[ 1 : midpnt ];
                        maxlags = maxlags, lookahead = lookahead, minlength = minlength, mincorr = mincorr);
    RHS = RecursiveAlignment(aligned[ (midpnt + 1) : end ], reference[ (midpnt + 1) : end ];
                        maxlags = maxlags, lookahead = lookahead, minlength = minlength, mincorr = mincorr);
    return vcat( LHS, RHS )
end

"""
    RAFFT(raw, reference; maxlags::Int = 500, lookahead::Int = 1, minlength::Int = 20, mincorr::Float64 = 0.05)

RAFFT corrects shifts in the `raw` spectral bands to be similar to those in a given `reference` spectra through
the use of "recursive alignment by FFT". It returns an array of corrected spectra/chromatograms. The number of maximum lags can be
specified, the `lookahead` parameter ensures that additional recursive executions are performed so the first solution
found is not preemptively accepted, the minimum segment length(`minlength`) can also be specified if FWHM are estimable,
and the minimum cross correlation(`mincorr`) for a match can dictate whether peaks were found to align or not.

*Note* This method works best with flat baselines because it repeats last known values when padding aligned spectra.
It is highly efficient, and in my tests does a good job, but other methods definitely exist. Let me know if other peak Alignment
methods are important for your work-flow, I'll see if I can implement them.

Application of Fast Fourier Transform Cross-Correlation for the Alignment of Large Chromatographic and Spectral Datasets
Jason W. H. Wong, Caterina Durante, and, Hugh M. Cartwright. Analytical Chemistry 2005 77 (17), 5655-5661
"""
function RAFFT(raw, reference; maxlags::Int = 500, lookahead::Int = 1, minlength::Int = 20, mincorr::Float64 = 0.05)
    corrected = zeros( size( raw ) )
    for sample in 1 : size( raw )[ 1 ]
        corrected[ sample, : ] = RecursiveAlignment( raw[ sample, : ], reference;
                                                    maxlags = maxlags,
                                                    lookahead = lookahead,
                                                    minlength = minlength,
                                                    mincorr = mincorr)
    end
    return corrected
end

struct COW
   Nodes::Array
   Lengths::Array
end

"""
    COW( A, B; segments = 20, slack = 1,
               maxslack = Int( floor( length(A) / segments ) ) - 2 )

COW makes a CorrelationOptimizedWarping object which warp corrects spectra `A`
to reference `B`. The user can select the number of segments, slack size,
and optionally the maximum slack parameter.

Note: Not fully tested.

"Aligning of single and multiple wavelength chromatographic profiles for chemometric data analysis using correlation optimised warping"
Nielsen, N. P. V.; Carstensen, J. M.; Smedsgaard, J.Journal of Chromatography A. 1998,805, 17–35.
"""
function COW( A, B;
                  segments = 20, slack = 1,
                  maxslack = Int( floor( length(A) / segments ) ) - 2 )
   @warn "This code (COW) is mostly untested. Please report any bugs!"
   vars = length( A )
   seglen = Int( floor( vars / segments ) )
   #@assert( vars == length( B ), "spectra should have the same dimensionality" )
   @assert(seglen > slack, "Slack cannot exceed the length of segments ($seglen)")
   @assert(maxslack < (seglen - 2), "Maximum slack cannot exceed the length of segments ($seglen - 2)")
   if seglen < 10
      @warn "the segment length is less then 10 bins. This could be a bad idea!"
   end
   seginit = collect(1 : seglen : vars)
   seginit[end] = vars
   extent = ( (length(seginit) - segments) == 1 ) ? (1:segments) : (1:(segments - 1))
   goalsegments = [ B[ seginit[i]:seginit[(i + 1)] ] for i in extent ]
   goallengths = length.(goalsegments)
   slacks = (-slack):slack
   nodes = copy(seginit)
   corr, oldstate = zeros( segments -1 ), zeros(Int, segments - 2 )
   highscore = -1.1 * segments
   bestslacks = zeros( segments-2 )
   iterate = Iterators.product( repeat( [ slacks ], segments - 2 )... )
   for i in iterate
      #Check change
      curstate = reverse( i ) #The reverse is easier to think about for me...
      difference = findfirst( oldstate .!= curstate )
      #Compute state
      update = seginit[(difference+1):(end-1)] .+ curstate[difference:end]
      nodes[ (difference+1):(end-1) ] = collect(update)
      #Check validity of segments
      if all(nodes[2:(end-1)] .<= vars) && all( abs.(nodes[2:(end-1)] .- seginit[2:(end-1)]) .<= maxslack )
         #Get score
         for d in difference : ( segments - 1 )
            span = nodes[ d ]: (nodes[ d + 1 ] - 1)
            resampled = LinearResample( A[ span ], goallengths[d] )
            corr[ d ] = Statistics.cor( resampled, goalsegments[d] )
         end
         #Store the best score
         if sum(corr) > highscore
            highscore = sum(corr)
            bestslacks = curstate
         end
      end
      #Update state
      oldstate = curstate
   end
   #Warp to best state
   warped = copy( A )
   update = seginit[2:(segments-1)] .+  bestslacks
   nodes = copy(seginit)
   nodes[ 2:(segments-1) ] = update
   return COW(nodes, length.(goalsegments) )
end

"""
    (moo::COW)(A)

Applies a learned warping from a COW object to spectrum A.
"""
function (moo::COW)(A)
   warped = []
   for s in 1 : (length(moo.Nodes) - 1)
      if s == 1
         span = 1 : (moo.Nodes[ s + 1 ] - 1 )
      else
         span = (moo.Nodes[ s ]+1) : (moo.Nodes[ s + 1 ] - 1 )
      end
      fudge = (s == (length(moo.Nodes) - 1)) ? -1 : 0
      warped = vcat(warped, LinearResample( A[ span ], moo.Lengths[ s ] + fudge ) )
   end
   return warped
end

"""
    DynamicTimeWarping( A, B )

Applies Dynamic Time Warping of spectrum `A` so it maps to reference `B`.

Returns the DTW distance, shortest path, and the DTW cost matrix. This does not
automatically rescale `A` to `B`, the way inwhich you want to do that is up to the user
for now.

Sakoe, Hiroaki; Chiba, Seibi (1978). "Dynamic programming algorithm optimization for spoken word recognition".
IEEE Transactions on Acoustics, Speech, and Signal Processing. 26 (1): 43–49. doi:10.1109/tassp.1978.1163055.
"""
function DynamicTimeWarping( A, B )
    #TODO: Add different distances
    DTW = SquareEuclideanDistance( A, B )
    Na, Nb = size( DTW )
    DTW[1,:] = cumsum(DTW[1,:])
    DTW[:,1] = cumsum(DTW[:,1])
    for i = 2:Na, j = 2:Nb
        options = DTW[i-1, j-1]
        if i < Na
            options = vcat( options, DTW[i-1, j] )
        end
        if j < Nb
            options = vcat( options, DTW[i, j-1] )
        end
        DTW[i, j] += minimum(options)
    end
    shortestpath = [(Na, Nb)]
    cost = 0.0
    for step in 2 : Na
        lastposition = shortestpath[end]
        if lastposition[1] == 1
            push!(shortestpath, (1, lastposition[2] - 1) )
        elseif lastposition[1] == 1
            push!(shortestpath, (lastposition[1] - 1, 1) )
        else
            legalmoves = [ lastposition .- 1,
                            (lastposition[1], lastposition[2] - 1),
                            (lastposition[1] - 1, lastposition[2]) ]
            bestmove = argmin( DTW[CartesianIndex.( legalmoves )] )
            push!(shortestpath, legalmoves[bestmove] )
        end
    end
    if shortestpath[end] != (1,1)
        push!(shortestpath, (1,1))
    end
    cost = sum( DTW[ CartesianIndex.( shortestpath ) ] )
    return cost, shortestpath, DTW
end

"""
    AssessHealth( X )

Returns a somewhat detailed Dict containing information about the 'health' of a dataset. What is included is the following:
    - PercentMissing: percent of missing entries (includes nothing, inf / nan) in the dataset
    - EmptyColumns: the columns which have only 1 value
    - RankEstimate: An estimate of the rank of X
    - (optional)Duplicates: returns the rows of duplicate observations
"""
function AssessHealth( X; checkduplicates = true )
    (obs, vars) = size( X )
    NumberMissing = sum( map( x -> isa(x, Missing) || isa(x, Nothing) || isnan(x) || isinf(x), X ) )
    PercentMissing = NumberMissing / (obs * vars)
    empties = []
    for c in 1 : vars
        if length(unique(X[:,c])) == 1
            push!(empties, c)
        end
    end
    RankEst = (NumberMissing == 0) ? LinearAlgebra.rank(X) : missing
    InfoDict = Dict("PercentMissing" => PercentMissing,
                    "EmptyColumns" => empties,
                    "RankEstimate" => RankEst)
    if checkduplicates
        if NumberMissing == 0
            dupes = []
            for o1 in 1:obs
                for o2 in (o1 + 1):obs
                    if X[o1,:] == X[o2,:]
                        push!(dupes, o1)
                    end
                end
            end
            InfoDict["Duplicates"] = dupes
        else
            InfoDict["Duplicates"] = missing
        end
    end
    return InfoDict
end
